from dataclasses import dataclass

from bs4 import BeautifulSoup

from kabutobashi.domain.entity import Weeks52HighLow
from .page import Page, PageDecoder


@dataclass(frozen=True)
class Weeks52HighLowPage(Page):
    data_type: str
    base_url: str = "https://jp.tradingview.com/markets/stocks-japan"

    def __post_init__(self):
        if self.data_type not in ["high", "low", "newly_high", "newly_low"]:
            raise ValueError()

    def url(self) -> str:
        prefix = None
        # 52週の高値・底値を取得する関数とURL
        if self.data_type == "high":
            prefix = "highs-and-lows-52wk-high"
        elif self.data_type == "low":
            prefix = "highs-and-lows-52wk-low"
        elif self.data_type == "newly_high":
            prefix = "highs-and-lows-ath"
        elif self.data_type == "newly_low":
            prefix = "highs-and-lows-atl"

        if prefix is None:
            raise ValueError()

        return f"{self.base_url}/{prefix}"

    @staticmethod
    def _crawl_volatility_info(table: BeautifulSoup, volatility_up_tag: str, volatility_down_tag: str) -> dict:
        # 上がったか、下がったかを判断するリスト
        evaluate_list = list()
        evaluate_list.extend(table.find_all("td", class_=volatility_up_tag))
        evaluate_list.extend(table.find_all("td", class_=volatility_down_tag))
        return Weeks52HighLowPage._volatility_replace_method(volatility_list=evaluate_list)

    @staticmethod
    def _volatility_replace_method(volatility_list: list) -> dict:
        """
        volatility_listにはweb pageから取得したデータが入っている
        suffixには高値か底値かを表す接頭辞を表す
        """
        if len(volatility_list) == 0:
            return {}
        ratio_candidate = volatility_list[0].text.replace("%", "")
        value_candidate = volatility_list[1].text
        return {"volatility_ratio": ratio_candidate, "volatility_value": value_candidate}

    def _get(self) -> dict:
        res = BeautifulSoup(self.get_url_text(url=self.url()), "lxml")

        # ここからcrawlするページのタグ
        cell = "tv-screener-table__cell"
        signal = "tv-screener-table__signal"
        close_tag = f"tv-data-table__cell {cell} {cell}--big"
        # タグ
        volatility_up_tag = f"{close_tag} {cell}--up"
        volatility_down_tag = f"{close_tag} {cell}--down"
        buy_tag = f"{signal} {signal}--buy"
        strong_buy_tag = f"{signal} {signal}--strong-buy"
        sell_tag = f"{signal} {signal}--sell"
        strong_sell_tag = f"{signal} {signal}--strong-sell"

        content = res.find("tbody", class_="tv-data-table__tbody")
        table = content.find_all("tr")
        whole_result = []
        for t in table:
            volatility_dict = self._crawl_volatility_info(
                table=t, volatility_up_tag=volatility_up_tag, volatility_down_tag=volatility_down_tag
            )

            data = {
                "code": PageDecoder(tag1="a").decode(bs=t),
                "brand_name": PageDecoder(tag1="span").decode(bs=t),
                "close": PageDecoder(tag1="td", class1=close_tag).decode(bs=t),
                "volatility_up": PageDecoder(tag1="td", class1=volatility_up_tag).decode(bs=t),
                "volatility_down": PageDecoder(tag1="td", class1=volatility_down_tag).decode(bs=t),
                "buy": PageDecoder(tag1="span", class1=buy_tag).decode(bs=t),
                "strong_buy": PageDecoder(tag1="span", class1=strong_buy_tag).decode(bs=t),
                "sell": PageDecoder(tag1="span", class1=sell_tag).decode(bs=t),
                "strong_sell": PageDecoder(tag1="span", class1=strong_sell_tag).decode(bs=t),
                "volatility_ratio": volatility_dict.get("volatility_ratio", "-"),
                "volatility_value": volatility_dict.get("volatility_value", "-"),
            }
            whole_result.append(Weeks52HighLow.from_page_of(data=data).dumps())

        return {"weeks_52_high_low": whole_result}



from dataclasses import dataclass
import matplotlib.pyplot as plt
import pandas as pd
from .method import Method


@dataclass(frozen=True)
class SMA(Method):
    short_term: int = 5
    medium_term: int = 21
    long_term: int = 70
    method_name: str = "sma"

    def _method(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df = _df.assign(
            sma_short=_df["close"].rolling(self.short_term).mean(),
            sma_medium=_df["close"].rolling(self.medium_term).mean(),
            sma_long=_df["close"].rolling(self.long_term).mean(),
        )
        return _df

    def _signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df["diff"] = _df.apply(lambda x: x["sma_long"] - x["sma_short"], axis=1)
        # 正負が交差した点
        _df = _df.join(self._cross(_df["diff"]))
        _df = _df.rename(columns={"to_plus": "buy_signal", "to_minus": "sell_signal"})
        return _df

    def _visualize(self, _df: pd.DataFrame):
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(6, 5))
        # x軸のオートフォーマット
        fig.autofmt_xdate()

        # set candlestick
        self.add_ax_candlestick(ax, _df)

        # plot macd
        ax.plot(_df.index, _df["sma_long"], color="#dc143c", label="sma_long")
        ax.plot(_df.index, _df["sma_medium"], color="#ffa500", label="sma_medium")
        ax.plot(_df.index, _df["sma_short"], color="#1e90ff", label="sma_short")

        ax.legend(loc="best")  # 各線のラベルを表示
        return fig



from dataclasses import dataclass
import pandas as pd
import matplotlib.pyplot as plt
from .method import Method


@dataclass(frozen=True)
class Ichimoku(Method):
    """

    See Also:
        https://kabu.com/investment/guide/technical/04.html
    """

    short_term: int = 12
    medium_term: int = 26
    long_term: int = 52
    method_name: str = "ichimoku"

    def _method(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df = _df.assign(
            # 短期の線
            short_max=lambda x: x["close"].rolling(self.short_term).max(),
            short_min=lambda x: x["close"].rolling(self.short_term).min(),
            # 中期の線
            medium_max=lambda x: x["close"].rolling(self.medium_term).max(),
            medium_min=lambda x: x["close"].rolling(self.medium_term).min(),
            # 長期線
            long_max=lambda x: x["close"].rolling(self.long_term).max(),
            long_min=lambda x: x["close"].rolling(self.long_term).min(),
        )

        # 指標の計算
        _df = _df.assign(
            line_change=lambda x: (x["short_max"] + x["short_min"]) / 2,
            line_base=lambda x: (x["medium_max"] + x["medium_min"]) / 2,
            # 先行線
            proceding_span_1=lambda x: (x["line_change"] + x["line_base"]) / 2,
            proceding_span_2=lambda x: (x["long_max"] + x["long_min"]) / 2,
        )

        # 値のshift
        _df = _df.assign(
            proceding_span_1=_df["proceding_span_1"].shift(26),
            proceding_span_2=_df["proceding_span_2"].shift(26),
            delayed_span=_df["close"].shift(26),
        )
        return _df

    def _signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        return _df

    def _visualize(self, _df: pd.DataFrame):
        # TODO implement
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(6, 5))
        # x軸のオートフォーマット
        fig.autofmt_xdate()

        # set candlestick
        self.add_ax_candlestick(ax, _df)

        # plot macd
        ax.plot(_df.index, _df["sma_long"], color="#dc143c", label="sma_long")

        ax.legend(loc="best")  # 各線のラベルを表示
        return fig



from dataclasses import dataclass
import pandas as pd
from .method import Method
import matplotlib.pyplot as plt


@dataclass(frozen=True)
class ADX(Method):
    """
    相場のトレンドの強さを見るための指標である`ADX`を計算するクラス。

    以下の指標を計算するクラス

    * +DI: 株価の上昇の大きさ
    * -DI: 株価の下降の大きさ
    * ADX: 株価のトレンドの強さ
    * ADXR: ADXの単純移動平均線

    Args:
        term (int):
        adx_term (int):
        adxr_term (int):

    See Also:
        * https://www.sevendata.co.jp/shihyou/technical/dmi.html
        * https://www.sevendata.co.jp/shihyou/technical/adx.html

    """

    term: int = 14
    adx_term: int = 14
    adxr_term: int = 28
    method_name: str = "adx"

    @staticmethod
    def _true_range(x: pd.DataFrame):
        """

        Args:
            x (pd.DataFrame)

        Returns:
            maximum
        """
        current_high = x["high"]
        current_low = x["low"]
        prev_close = x["shift_close"]
        a = current_high - current_low
        b = current_high - prev_close
        c = prev_close - current_low
        max_ab = max(a, b)
        max_ac = max(a, c)
        return max(max_ab, max_ac)

    @staticmethod
    def _compute_dx(x: pd.DataFrame) -> float:
        """

        Args:
            x (pd.DataFrame):
        """
        numerator = abs(x["plus_di"] - x["minus_di"])
        denominator = x["plus_di"] + x["minus_di"]
        return numerator / denominator * 100

    @staticmethod
    def _fixed_plus_dm(x: pd.DataFrame) -> float:
        if x["plus_dm"] > 0 and x["plus_dm"] > x["minus_dm"]:
            return x["plus_dm"]
        else:
            return 0

    @staticmethod
    def _fixed_minus_dm(x: pd.DataFrame) -> float:
        if x["minus_dm"] > 0 and x["minus_dm"] > x["plus_dm"]:
            return x["minus_dm"]
        else:
            return 0

    def _method(self, _df: pd.DataFrame) -> pd.DataFrame:
        # 利用する値をshift
        _df = _df.assign(
            shift_high=_df["high"].shift(1), shift_low=_df["low"].shift(1), shift_close=_df["close"].shift(1)
        )
        _df = _df.assign(
            plus_dm=_df.apply(lambda x: x["high"] - x["shift_high"], axis=1),
            minus_dm=_df.apply(lambda x: x["shift_low"] - x["low"], axis=1),
        )
        _df = _df.assign(
            fixed_plus_dm=_df.apply(lambda x: self._fixed_plus_dm(x), axis=1),
            fixed_minus_dm=_df.apply(lambda x: self._fixed_minus_dm(x), axis=1),
        )
        _df = _df.assign(
            true_range=_df.apply(lambda x: self._true_range(x), axis=1),
            sum_tr=lambda x: x["true_range"].rolling(self.term).sum(),
            sum_plus_dm=lambda x: x["fixed_plus_dm"].rolling(self.term).sum(),
            sum_minus_dm=lambda x: x["fixed_minus_dm"].rolling(self.term).sum(),
        )

        _df = _df.dropna()
        required_columns = ["open", "high", "low", "close", "sum_plus_dm", "sum_minus_dm", "sum_tr"]
        _df = _df.loc[:, required_columns]

        # +DI, -DI
        _df = _df.assign(
            plus_di=_df.apply(lambda x: x["sum_plus_dm"] / x["sum_tr"] * 100, axis=1),
            minus_di=_df.apply(lambda x: x["sum_minus_dm"] / x["sum_tr"] * 100, axis=1),
        )
        _df = _df.assign(
            DX=_df.apply(self._compute_dx, axis=1),
            ADX=lambda x: x["DX"].rolling(self.adx_term).mean(),
            ADXR=lambda x: x["DX"].rolling(self.adxr_term).mean(),
        )
        return _df

    @staticmethod
    def _buy_signal(x) -> float:
        """
        DMIとADXを組み合わせた基本パターン
        """

        # +DIが-DIを上抜き、ADXが上昇傾向の上向きであれば新規買い
        if x["ADX_trend"] > 0:
            if x["to_plus"] > 0:
                return 1

        # +DIが-DIより上に位置している際に、
        # ADXが下向きから上向きに転換した場合
        # if x['diff'] > 0:

    @staticmethod
    def _sell_signal(x) -> float:
        """
        DMIとADXを組み合わせた基本パターン
        """
        # -DIが+DIを下抜き、ADXが下落傾向の下向きであれば新規空売り
        if x["ADX_trend"] < 0:
            if x["to_minus"] > 0:
                return 1

    def _signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        """
        buy_signalとsell_signalを付与
        """
        _df["ADX_trend"] = self._trend(_df["ADX"])
        _df["diff"] = _df["plus_di"] - _df["minus_di"]
        _df = _df.join(self._cross(_df["diff"]))

        _df["buy_signal"] = _df.apply(lambda x: self._buy_signal)
        _df["sell_signal"] = _df.apply(lambda x: self._sell_signal)

        return _df

    def _visualize(self, _df: pd.DataFrame):
        fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, gridspec_kw={"height_ratios": [3, 1]}, figsize=(6, 5))
        # x軸のオートフォーマット
        fig.autofmt_xdate()

        # set candlestick
        self.add_ax_candlestick(ax1, _df)

        # plot adx
        ax2.plot(_df.index, _df["plus_di"], label="+DI")
        ax2.plot(_df.index, _df["minus_di"], label="-DI")
        ax2.plot(_df.index, _df["ADX"], label="ADX")
        ax2.plot(_df.index, _df["ADXR"], label="ADXR")
        ax2.legend(loc="center left")  # 各線のラベルを表示

        ax1.legend(loc="best")  # 各線のラベルを表示
        return fig



"""
io module provides input/output method
"""

from datetime import datetime
from typing import Union, Optional

import pandas as pd


def example_data() -> pd.DataFrame:
    """
    get example stock data

    Returns:
        stock data
    """
    data_path_list = ["../data/stooq.csv"]
    _df = read_csv(data_path_list)
    _df = _df.sort_values("date", ascending=True)
    _df = _df.convert_dtypes()
    return _df


def read_csv(path_candidate: Union[str, list], **kwargs) -> Optional[pd.DataFrame]:
    """
    通常のread_csvの関数に加えて、strとlist[str]の場合に縦方向に結合してDataFrameを返す

    Args:
        path_candidate: "path" or ["path_1", "path_2"]

    Returns:
        株のDataFrame
    """
    if type(path_candidate) is str:
        return pd.read_csv(path_candidate, **kwargs)
    elif type(path_candidate) is list:
        if not path_candidate:
            return None
        df_list = [pd.read_csv(p, **kwargs) for p in path_candidate]
        return pd.concat(df_list)
    else:
        return None


def read_stock_csv(
    path_candidate: Union[str, list],
    code_list: Optional[list] = None,
    drop_reit: bool = True,
    row_more_than: Optional[int] = None,
    **kwargs
) -> Optional[pd.DataFrame]:
    """
    本APIにてCrawlしたデータを扱いやすい形式にデータ変換する関数

    Args:
        path_candidate: "path" or ["path_1", "path_2"]
        code_list: filter with code_list
        drop_reit: drop REIT-data if True
        row_more_than: filter specified code, which has {row_more_than} data

    Returns:
        株のDataFrame
    """
    df = read_csv(path_candidate, **kwargs)
    if df is None:
        return None
    else:
        decoded_df = _decode_stock_data(_df=df)
        if code_list:
            decoded_df = decoded_df[decoded_df["code"].isin(code_list)]
        if drop_reit:
            decoded_df = decoded_df[~(decoded_df["market"] == " 東証REIT")]
        if row_more_than:
            dt_count = decoded_df.loc[:, ["code", "dt"]].groupby("code").count().reset_index()
            dt_count = dt_count[dt_count["dt"] >= row_more_than]
            _code_list = list(dt_count["code"].values)
            decoded_df = decoded_df[decoded_df["code"].isin(_code_list)]
        return decoded_df


def _decode_stock_data(_df: pd.DataFrame) -> pd.DataFrame:
    """
    以下のような株のデータを扱いやすいように整形する関数
    stock_label,name,close,date,industry_type,open,high,low,unit,per,psr,pbr,volume,market_capitalization,issued_shares,crawl_datetime,code
    1436  東証マザーズ,フィット,540.0,株価(15:00),業種建設業,555.0円,557.0円,482.0円,100株,---,0.46倍,0.54倍,"22,000株","2,312百万円","4,282千株",2020-03-13T23:31:04,1436
    1438  名証２部,岐阜造園,"1,098.0",株価(12:38),業種建設業,"1,110.0円","1,150.0円","1,098.0円",100株,9.19倍,0.38倍,0.62倍,"3,100株","1,594百万円","1,451千株",2020-03-13T23:31:06,1438
    :param _df:
    :return:
    """

    # 正規表現を利用して数値のみにする
    _df = _df.assign(
        market=_df["stock_label"].str.extract("[0-9]+ (.+)", expand=False),
        open=_df["open"].str.extract("(.+)円", expand=False),
        high=_df["high"].str.extract("(.+)円", expand=False),
        low=_df["low"].str.extract("(.+)円", expand=False),
        unit=_df["unit"].str.extract("(.+)株", expand=False),
        per=_df["per"].str.extract("(.+)倍", expand=False),
        psr=_df["psr"].str.extract("(.+)倍", expand=False),
        pbr=_df["pbr"].str.extract("(.+)倍", expand=False),
        volume=_df["volume"].str.extract("(.+)株", expand=False),
        dt=_df["crawl_datetime"].apply(lambda x: datetime.strptime(x, "%Y-%m-%dT%H:%M:%S").strftime("%Y-%m-%d")),
    )

    # 必要なカラムに絞る
    required_columns = ["code", "open", "close", "high", "low", "unit", "volume", "per", "psr", "pbr", "market", "dt"]
    _df = _df.loc[:, required_columns].drop_duplicates()
    # 変な値はpd.NAに変換
    _df = _df.replace("---", pd.NA)
    return _df



from dataclasses import dataclass, asdict
from cerberus import Validator
from kabutobashi.errors import KabutobashiEntityError


@dataclass(frozen=True)
class StockInfo:
    """
    Args:
        code: 銘柄コード
        market: 市場
        industry_type: 業種
        open: 円
        high: 円
        low: 円
        close: 円
    """

    code: str
    market: str
    industry_type: str
    open: float
    high: float
    low: float
    close: float
    psr: float
    per: float
    pbr: float
    volume: int
    unit: int
    market_capitalization: str
    issued_shares: str
    dt: str
    _SCHEMA = {
        "code": {"type": "string"},
        "market": {"type": "string"},
        "industry_type": {"type": "string"},
        "open": {"type": "float"},
        "high": {"type": "float"},
        "low": {"type": "float"},
        "close": {"type": "float"},
        "psr": {"type": "float"},
        "per": {"type": "float"},
        "pbr": {"type": "float"},
        "volume": {"type": "integer"},
        "unit": {"type": "integer"},
        "market_capitalization": {"type": "string"},
        "issued_shares": {"type": "string"},
        "dt": {"type": "string"},
    }

    def __post_init__(self):
        validator = Validator(self._SCHEMA)
        if not validator.validate(self.dumps()):
            raise KabutobashiEntityError(validator)

    @staticmethod
    def from_page_of(data: dict) -> "StockInfo":
        label_split = data["stock_label"].split("  ")
        return StockInfo(
            code=label_split[0],
            market=label_split[1],
            industry_type=data["industry_type"],
            open=float(StockInfo._convert(data["open"])),
            high=float(StockInfo._convert(data["high"])),
            low=float(StockInfo._convert(data["low"])),
            close=float(StockInfo._convert(data["close"])),
            unit=int(StockInfo._convert(data["unit"])),
            psr=float(StockInfo._convert(data["psr"])),
            per=float(StockInfo._convert(data["per"])),
            pbr=float(StockInfo._convert(data["pbr"])),
            volume=int(StockInfo._convert(data["volume"])),
            market_capitalization=data["market_capitalization"],
            issued_shares=data["issued_shares"],
            dt=data["date"],
        )

    @staticmethod
    def _convert(input_value: str) -> str:
        if input_value == "---":
            return "0"
        return input_value.replace("円", "").replace("株", "").replace("倍", "").replace(",", "")

    def dumps(self) -> dict:
        return asdict(self)



from dataclasses import dataclass
import pandas as pd
import matplotlib.pyplot as plt
from .method import Method


@dataclass(frozen=True)
class BollingerBands(Method):
    """
    株価の勢いの変化や反転の目安、方向を見る`BollingerBands`を計算するクラス。

    See Also:
        * https://www.sevendata.co.jp/shihyou/technical/bori.html
    """

    band_term: int = 12
    continuity_term: int = 10
    method_name: str = "bollinger_bands"

    def _method(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df = _df.assign(
            mean=_df["close"].rolling(self.band_term).mean(), std=_df["close"].rolling(self.band_term).std()
        )
        _df = _df.assign(
            upper_1_sigma=_df.apply(lambda x: x["mean"] + x["std"] * 1, axis=1),
            lower_1_sigma=_df.apply(lambda x: x["mean"] - x["std"] * 1, axis=1),
            upper_2_sigma=_df.apply(lambda x: x["mean"] + x["std"] * 2, axis=1),
            lower_2_sigma=_df.apply(lambda x: x["mean"] - x["std"] * 2, axis=1),
            upper_3_sigma=_df.apply(lambda x: x["mean"] + x["std"] * 3, axis=1),
            lower_3_sigma=_df.apply(lambda x: x["mean"] - x["std"] * 3, axis=1),
        )
        return _df

    def _signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df = _df.assign(
            over_upper=_df.apply(lambda x: 1 if x["close"] > x["upper_2_sigma"] else 0, axis=1),
            over_lower=_df.apply(lambda x: 1 if x["close"] < x["lower_2_sigma"] else 0, axis=1),
            over_upper_continuity=lambda x: x["over_upper"].rolling(self.continuity_term).sum(),
            over_lower_continuity=lambda x: x["over_lower"].rolling(self.continuity_term).sum(),
        )

        _df["buy_signal"] = _df["over_upper"].apply(lambda x: 1 if x > 0 else 0)
        _df["sell_signal"] = _df["over_lower"].apply(lambda x: 1 if x > 0 else 0)
        return _df

    def _visualize(self, _df: pd.DataFrame):
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(6, 5))
        # x軸のオートフォーマット
        fig.autofmt_xdate()

        # set candlestick
        self.add_ax_candlestick(ax, _df)

        # plot
        ax.plot(_df.index, _df["upper_1_sigma"], color="#dc143c", label="+1s")
        ax.plot(_df.index, _df["lower_1_sigma"], color="#dc143c", label="-1s")
        ax.plot(_df.index, _df["upper_2_sigma"], color="#ffa500", label="+2s")
        ax.plot(_df.index, _df["lower_2_sigma"], color="#ffa500", label="-2s")
        ax.plot(_df.index, _df["upper_3_sigma"], color="#1e90ff", label="+3s")
        ax.plot(_df.index, _df["lower_3_sigma"], color="#1e90ff", label="-3s")

        ax.legend(loc="best")  # 各線のラベルを表示
        return fig



import pytest
import kabutobashi as kb


class TestStockInfo:

    def test_error_init(self):
        with pytest.raises(kb.errors.KabutobashiEntityError):
            _ = kb.StockInfo(
                code="1234",
                market="market",
                industry_type="industry_type",
                open="",
                high="",
                low="",
                close="",
                psr="",
                per="",
                pbr="",
                volume="",
                unit="",
                market_capitalization="",
                issued_shares="",
                dt=""
            )


class TestStockIpo:

    def test_error_init(self):
        with pytest.raises(kb.errors.KabutobashiEntityError):
            _ = kb.StockIpo(
                code="",
                market="",
                manager="",
                stock_listing_at="",
                public_offering="",
                evaluation="",
                initial_price=""
            )


class TestWeeks52HihLow:

    def test_error_init(self):
        with pytest.raises(kb.errors.KabutobashiEntityError):
            _ = kb.Weeks52HighLow(
                code="",
                brand_name="",
                close="",
                buy_or_sell="",
                volatility_ratio="",
                volatility_value=""
            )



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from functools import reduce
from typing import List, Union

from bs4 import BeautifulSoup
from kabutobashi.errors import KabutobashiPageError
import requests

from .user_agent import UserAgent


@dataclass(frozen=True)
class PageDecoder:
    tag1: str = None
    class1: str = None
    id1: str = None
    default: str = ""

    def _decode(self, value):
        class1 = {"class": self.class1}

        set_value = None
        # tag1から取得
        if self.tag1 is not None:
            if class1["class"] is not None:
                set_value = value.find(self.tag1, self.class1)
            else:
                set_value = value.find(self.tag1)

        if set_value is None:
            return self.default

        # 文字列を置換して保持
        return self.replace(set_value.get_text())

    def decode(self, bs: BeautifulSoup) -> Union[str, List[str]]:
        return self._decode(value=bs)

    @staticmethod
    def replace(input_text: str) -> str:
        target_list = [" ", "\t", "\n", "\r", "円"]

        def remove_of(_input: str, target: str):
            return _input.replace(target, "")

        result = reduce(remove_of, target_list, input_text)
        return result.replace("\xa0", " ")


@dataclass(frozen=True)
class Page(metaclass=ABCMeta):
    @abstractmethod
    def url(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def _get(self) -> dict:
        raise NotImplementedError()

    def get(self) -> dict:
        data = self._get()
        data.update({"crawl_datetime": self.crawl_datetime()})
        return data

    @staticmethod
    def get_url_text(url: str) -> str:
        """
        requestsを使って、webからページを取得し、htmlを返す
        """
        user_agent = UserAgent.get_user_agent_header()
        r = requests.get(url, headers=user_agent)

        if r.status_code != 200:
            raise KabutobashiPageError(url=url)

        # 日本語に対応
        r.encoding = r.apparent_encoding
        return r.text

    @staticmethod
    def crawl_datetime() -> str:
        jst = timezone(timedelta(hours=+9), "JST")
        now = datetime.now(jst)
        return now.strftime("%Y-%m-%dT%H:%M:%S")



from dataclasses import dataclass
import matplotlib.pyplot as plt
import pandas as pd
from .method import Method


@dataclass(frozen=True)
class Momentum(Method):
    """
    See Also:
        https://www.sevendata.co.jp/shihyou/technical/momentum.html
    """

    term: int = 25
    method_name: str = "momentum"

    def _method(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df = _df.assign(
            momentum=_df["close"].shift(10), sma_momentum=lambda x: x["momentum"].rolling(self.term).mean()
        )
        return _df

    def _signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df = _df.join(self._cross(_df["sma_momentum"]))
        _df = _df.rename(columns={"to_plus": "buy_signal", "to_minus": "sell_signal"})
        return _df

    def _visualize(self, _df: pd.DataFrame):
        fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, gridspec_kw={"height_ratios": [3, 1]}, figsize=(6, 5))
        # x軸のオートフォーマット
        fig.autofmt_xdate()

        # set candlestick
        self.add_ax_candlestick(ax1, _df)

        # plot macd
        ax2.plot(_df.index, _df["momentum"], label="momentum")
        ax2.plot(_df.index, _df["sma_momentum"], label="sma_momentum")
        ax2.legend(loc="center left")  # 各線のラベルを表示

        ax1.legend(loc="best")  # 各線のラベルを表示
        return fig



"""
Method modules provide technical analysis for stock chart.


"""
from .method import Method
from .sma import SMA
from .macd import MACD
from .stochastics import Stochastics
from .adx import ADX
from .bollinger_bands import BollingerBands
from .ichimoku import Ichimoku
from .momentum import Momentum
from .psycho_logical import PsychoLogical
from .fitting import Fitting



from dataclasses import dataclass
import matplotlib.pyplot as plt
import pandas as pd
from .method import Method


@dataclass(frozen=True)
class MACD(Method):
    """ """

    short_term: int = 12
    long_term: int = 26
    macd_span: int = 9
    method_name: str = "macd"

    def _method(self, _df: pd.DataFrame) -> pd.DataFrame:
        """
        macdを基準として今後上昇するかどうかをスコアで返す。
        値が大きければその傾向が高いことを表している。
        最小値は0で、最大値は無限大である。
        :param _df:
        :return:
        """
        # histogramが図として表現されるMACDの値
        _df = _df.assign(
            # MACDの計算
            ema_short=lambda x: x["close"].ewm(span=self.short_term).mean(),
            ema_long=lambda x: x["close"].ewm(span=self.long_term).mean(),
            macd=lambda x: x["ema_short"] - x["ema_long"],
            signal=lambda x: x["macd"].ewm(span=self.macd_span).mean(),
            # ヒストグラム値
            histogram=lambda x: x["macd"] - x["signal"],
        )
        return _df

    def _signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        # 正負が交差した点
        _df = _df.join(self._cross(_df["histogram"]))
        _df = _df.rename(columns={"to_plus": "buy_signal", "to_minus": "sell_signal"})
        return _df

    def _visualize(self, _df: pd.DataFrame):
        fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, gridspec_kw={"height_ratios": [3, 1]}, figsize=(6, 5))
        # x軸のオートフォーマット
        fig.autofmt_xdate()

        # set candlestick
        self.add_ax_candlestick(ax1, _df)

        # plot macd
        ax2.plot(_df.index, _df["macd"], label="macd")
        ax2.plot(_df.index, _df["signal"], label="signal")
        ax2.bar(_df.index, _df["histogram"], label="histogram")
        ax2.legend(loc="center left")  # 各線のラベルを表示

        ax1.legend(loc="best")  # 各線のラベルを表示
        return fig



from dataclasses import dataclass
import pandas as pd
from scipy.optimize import curve_fit
import numpy as np
import matplotlib.pyplot as plt
from .method import Method


@dataclass(frozen=True)
class Fitting(Method):
    """ """

    def __init__(self):
        super().__init__(method_name="macd")

    def _method(self, _df: pd.DataFrame) -> pd.DataFrame:
        """
        macdを基準として今後上昇するかどうかをスコアで返す。
        値が大きければその傾向が高いことを表している。
        最小値は0で、最大値は無限大である。
        :param _df:
        :return:
        """
        # histogramが図として表現されるMACDの値
        array_y = _df["close"]
        array_x = np.array(range(0, len(array_y)))

        def _linear_fit(x, a, b):
            return a * x + b

        def _square_fit(x, a, b, c):
            return a * x * x + b * x + c

        def _cube_fit(x, a, b, c, d):
            return a * x * x * x + b * x * x + c * x + d

        linear_param, _ = curve_fit(_linear_fit, array_x, array_y)
        square_param, _ = curve_fit(_square_fit, array_x, array_y)
        cube_param, _ = curve_fit(_cube_fit, array_x, array_y)
        _df["linear_fitting"] = [_linear_fit(x, *linear_param) for x in array_x]
        _df["square_fitting"] = [_square_fit(x, *square_param) for x in array_x]
        _df["cube_fitting"] = [_cube_fit(x, *cube_param) for x in array_x]
        return _df

    def _signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        return _df

    def _visualize(self, _df: pd.DataFrame):
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(6, 5))
        # x軸のオートフォーマット
        fig.autofmt_xdate()

        # set candlestick
        self.add_ax_candlestick(ax, _df)

        # plot
        ax.plot(_df.index, _df["linear_fitting"], color="#dc143c", label="linear")
        ax.plot(_df.index, _df["square_fitting"], color="#ffa500", label="square")
        ax.plot(_df.index, _df["cube_fitting"], color="#1e90ff", label="cube")

        ax.legend(loc="best")  # 各線のラベルを表示
        return fig



from datetime import datetime, timedelta
import jpholiday
import numpy as np
import pandas as pd
from scipy.optimize import curve_fit

from .errors import KabutobashiBaseError, StockDfError
from kabutobashi.domain.method import *


def get_past_n_days(current_date: str, n: int = 60) -> list:
    """
    土日と祝日を考慮したn営業日前までの日付のリストを返す関数

    Args:
        current_date: n日前を計算する起点となる日
        n: n日前

    Returns:
        date list, ex ["%Y-%m-%d", "%Y-%m-%d", "%Y-%m-%d", ...]
    """
    multiply_list = [2, 4, 8, 16]
    for multiply in multiply_list:
        return_candidate = _get_past_n_days(current_date=current_date, n=n, multiply=multiply)
        if len(return_candidate) == n:
            return return_candidate
    raise KabutobashiBaseError(f"{n}日前を正しく取得できませんでした")


def _get_past_n_days(current_date: str, n: int, multiply: int) -> list:
    """
    n*multiplyの日数分のうち、商取引が行われる日を取得する

    Args:
        current_date: n日前を計算する起点となる日
        n: n日前
        multiply: n日前にかける数。
    """
    end_date = datetime.strptime(current_date, "%Y-%m-%d")
    # 2倍しているのは土日や祝日が排除されるため
    # また、nが小さすぎると休日が重なった場合に日数の取得ができないため
    back_n_days = n * multiply
    date_candidate = [end_date - timedelta(days=d) for d in range(back_n_days)]
    # 土日を除く
    filter_weekend = [d for d in date_candidate if d.weekday() < 5]
    # 祝日を除く
    filter_holiday = [d for d in filter_weekend if not jpholiday.is_holiday(d)]
    # 文字列に日付を変えてreturn
    return list(map(lambda x: x.strftime("%Y-%m-%d"), filter_holiday[:n]))


def iter_by_code(stock_df: pd.DataFrame, days_thresholds: int = 60) -> (int, pd.DataFrame):
    """
    銘柄コードでイテレーションを回しつつ、必要なデータ数がある銘柄のDataFrameのみを返す関数。

    Args:
        stock_df:
        days_thresholds:

    Returns:
        (code: int, _df: pd.DataFrame)
    """
    for code, _df in stock_df.groupby("code"):
        if len(_df.index) < days_thresholds:
            continue
        else:
            yield code, format_to_stock_df(_df)


def replace_comma(x) -> float:
    """
    pandas内の値がカンマ付きの場合に、カンマを削除する関数

    Args:
        x:

    Returns:

    """
    if type(x) is str:
        x = x.replace(",", "")
    try:
        f = float(x)
    except ValueError:
        raise StockDfError(f"floatに変換できる値ではありません。")
    return f


def format_to_stock_df(_df: pd.DataFrame) -> pd.DataFrame:
    stock_df = (
        _df.replace("---", np.nan)
        .dropna(subset=["open", "close", "high", "low"])
        .fillna(0)
        .assign(
            open=_df["open"].apply(replace_comma),
            close=_df["close"].apply(replace_comma),
            high=_df["high"].apply(replace_comma),
            low=_df["low"].apply(replace_comma),
            volume=_df["volume"].apply(replace_comma),
        )
        .convert_dtypes()
    )
    return stock_df


def train_test_sliding_split(
    stock_df: pd.DataFrame, *, buy_sell_term_days: int = 5, sliding_window: int = 60, step: int = 2
):
    """

    Args:
        stock_df:
        buy_sell_term_days:
        sliding_window:
        step:

    Returns:

    """
    df_length = len(stock_df.index)
    if df_length < buy_sell_term_days + sliding_window:
        raise StockDfError("入力されたDataFrameの長さがwindow幅よりも小さいです")
    loop = df_length - (buy_sell_term_days + sliding_window)
    for idx, i in enumerate(range(0, loop, step)):
        offset = i + sliding_window
        yield idx, stock_df[i:offset], stock_df[offset : offset + buy_sell_term_days]


def compute_fitting(array_y: list, prefix: str) -> dict:
    statistical_values = {}
    array_x = np.array(range(0, len(array_y)))

    def _linear_fit(x, a, b):
        return a * x + b

    def _square_fit(x, a, b, c):
        return a * x * x + b * x + c

    def _cube_fit(x, a, b, c, d):
        return a * x * x * x + b * x * x + c * x + d

    linear_param, _ = curve_fit(_linear_fit, array_x, array_y)
    statistical_values.update({f"{prefix}_linear_{idx}": p for idx, p in enumerate(linear_param[:-1])})
    square_param, _ = curve_fit(_square_fit, array_x, array_y)
    statistical_values.update({f"{prefix}_square_{idx}": p for idx, p in enumerate(square_param[:-1])})
    cube_param, _ = curve_fit(_cube_fit, array_x, array_y)
    statistical_values.update({f"{prefix}_cube_{idx}": p for idx, p in enumerate(cube_param[:-1])})
    return statistical_values


def compute_statistical_values(stock_df: pd.DataFrame, fitting_term: int = 10, taken: int = 5) -> dict:
    # create and initialize instance
    _sma = SMA(short_term=5, medium_term=21, long_term=70)
    _macd = MACD(short_term=12, long_term=26, macd_span=9)
    _adx = ADX()
    _stochastics = Stochastics()

    statistical_values = {}
    statistical_values.update(compute_fitting(stock_df["close"], prefix="close"))
    # params with SMA0
    sma_df = stock_df.pipe(_sma)
    statistical_values.update(compute_fitting(sma_df["sma_short"][-fitting_term:], "sma_short"))
    statistical_values.update(compute_fitting(sma_df["sma_medium"][-fitting_term:], "sma_medium"))
    statistical_values.update(compute_fitting(sma_df["sma_long"][-fitting_term:], "sma_long"))
    # params with MACD
    macd_df = stock_df.pipe(_macd)
    statistical_values.update({f"histogram_{idx}": v for idx, v in enumerate(macd_df[-taken:]["histogram"].values)})
    adx_df = stock_df.pipe(_adx)
    statistical_values.update({f"plus_di_{idx}": v for idx, v in enumerate(adx_df[-taken:]["plus_di"].values)})
    statistical_values.update({f"minus_di_{idx}": v for idx, v in enumerate(adx_df[-taken:]["minus_di"].values)})
    statistical_values.update({f"ADX_{idx}": v for idx, v in enumerate(adx_df[-taken:]["ADX"].values)})
    statistical_values.update({f"ADXR_{idx}": v for idx, v in enumerate(adx_df[-taken:]["ADXR"].values)})
    # params with Stochastics
    stochastics_df = stock_df.pipe(_stochastics)
    statistical_values.update({f"D_{idx}": v for idx, v in enumerate(stochastics_df[-taken:]["D"].values)})
    statistical_values.update({f"SD_{idx}": v for idx, v in enumerate(stochastics_df[-taken:]["SD"].values)})

    return statistical_values



from dataclasses import dataclass
import pandas as pd
from kabutobashi.errors import StockDfError


@dataclass(frozen=True)
class StockDf:
    data_df: pd.DataFrame
    REQUIRED_COL = ["open", "high", "low", "close"]

    def __post_init__(self):
        self._null_check()
        self._code_constraint_check()

    def _null_check(self):
        if self.data_df is None:
            raise StockDfError("required")

    def _code_constraint_check(self):
        df_columns = self.data_df.columns
        if "code" in df_columns:
            code = list(set(self.data_df.code.values))
            if len(code) > 1:
                raise StockDfError("multiple code")
            elif len(code) == 0:
                raise StockDfError("no code")

    def _required_column_check(self):
        columns = list(self.data_df.columns)
        # 必須のカラム確認
        if not all([item in columns for item in self.REQUIRED_COL]):
            raise StockDfError(f"required: {self.REQUIRED_COL}, input: {columns}")

    @staticmethod
    def of(df: pd.DataFrame):
        df_columns = df.columns
        # 日付カラムの候補値を探す
        date_column = None
        if "date" in df_columns:
            date_column = "date"
        elif "dt" in df_columns:
            date_column = "dt"
        if date_column is None:
            raise StockDfError("日付のカラム[dt, date]のいずれかが存在しません")
        if "date" in df_columns and "dt" in df_columns:
            raise StockDfError("日付のカラム[dt, date]は片方しか存在できません")

        # indexにdateを指定
        df.index = pd.to_datetime(df[date_column]).sort_index()

        # 必要なカラムに絞る
        df = df[StockDf.REQUIRED_COL]
        open_s = df["open"].apply(StockDf._replace_comma)
        close_s = df["close"].apply(StockDf._replace_comma)
        high_s = df["high"].apply(StockDf._replace_comma)
        low_s = df["low"].apply(StockDf._replace_comma)
        return StockDf(data_df=pd.DataFrame({"open": open_s, "high": high_s, "low": low_s, "close": close_s}))

    @staticmethod
    def _replace_comma(x) -> float:
        """
        pandas内の値がカンマ付きの場合に、カンマを削除する関数
        :param x:
        :return:
        """
        if type(x) is str:
            x = x.replace(",", "")
        try:
            f = float(x)
        except ValueError as e:
            raise StockDfError(f"floatに変換できる値ではありません。{e}")
        return f



from dataclasses import dataclass
import pandas as pd
import matplotlib.pyplot as plt
from .method import Method


@dataclass(frozen=True)
class PsychoLogical(Method):
    """
    See Also:
        https://www.sevendata.co.jp/shihyou/technical/psycho.html
    """

    upper_threshold: float = 0.75
    lower_threshold: float = 0.25
    psycho_term: float = 12
    method_name: str = "psycho_logical"

    def _method(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df["shift_close"] = _df["close"].shift(1)
        _df["diff"] = _df.apply(lambda x: x["close"] - x["shift_close"], axis=1)

        _df["is_raise"] = _df["diff"].apply(lambda x: 1 if x > 0 else 0)

        _df["psycho_sum"] = _df["is_raise"].rolling(self.psycho_term).sum()
        _df["psycho_line"] = _df["psycho_sum"].apply(lambda x: x / self.psycho_term)

        _df["bought_too_much"] = _df["psycho_line"].apply(lambda x: 1 if x > self.upper_threshold else 0)
        _df["sold_too_much"] = _df["psycho_line"].apply(lambda x: 1 if x < self.lower_threshold else 0)
        return _df

    def _signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df["buy_signal"] = _df["sold_too_much"]
        _df["sell_signal"] = _df["bought_too_much"]
        return _df

    def _visualize(self, _df: pd.DataFrame):
        fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, gridspec_kw={"height_ratios": [3, 1]}, figsize=(6, 5))
        # x軸のオートフォーマット
        fig.autofmt_xdate()

        # set candlestick
        self.add_ax_candlestick(ax1, _df)

        # plot
        ax2.plot(_df.index, _df["psycho_line"], label="psycho_line")
        ax2.legend(loc="center left")  # 各線のラベルを表示

        ax1.legend(loc="best")  # 各線のラベルを表示
        return fig



from dataclasses import dataclass, asdict, field
from cerberus import Validator
from kabutobashi.errors import KabutobashiEntityError


@dataclass(frozen=True)
class StockIpo:
    """
    まだ取り込んでない値など

    '想定(仮条件)': '1,920(1,900-1,920)',
    '吸収金額': '75.6億',
    '(騰落率)損益': '(+1.1%)+2,100円00001',

    """

    code: int
    market: str = field(metadata={"jp": "市場"})
    manager: str = field(metadata={"jp": "主幹"})
    stock_listing_at: str = field(metadata={"jp": "上場日"})
    public_offering: float = field(metadata={"jp": "公募"})
    evaluation: str = field(metadata={"jp": "評価"})
    initial_price: float = field(metadata={"jp": "初値"})
    # current_price: float = field(metadata={"jp": "現在値"})
    # diff_price: float = field(metadata={"jp": "差分"})
    _SCHEMA = {
        "code": {"type": "string"},
        "market": {"type": "string"},
        "manager": {"type": "string"},
        "stock_listing_at": {"type": "string"},
        "public_offering": {"type": "float"},
        "evaluation": {"type": "string"},
        "initial_price": {"type": "float"},
    }

    def __post_init__(self):
        validator = Validator(self._SCHEMA)
        if not validator.validate(self.dumps()):
            raise KabutobashiEntityError(validator)

    @staticmethod
    def from_page_of(data: dict) -> "StockCode":
        # current_split = data["現在値(差分)"].split("(")
        return StockIpo(
            code=data["code"],
            market=data["市場"],
            manager=data["主幹"],
            stock_listing_at=data["上場"],
            public_offering=float(StockIpo._convert(data["公募"])),
            evaluation=data["評価"],
            initial_price=float(StockIpo._convert(data["初値"])),
            # current_price=0,
            # diff_price=0
        )

    @staticmethod
    def _convert(input_value: str) -> str:
        if input_value == "-":
            return "0"
        return input_value.replace("円", "").replace("株", "").replace("倍", "").replace(",", "")

    def dumps(self) -> dict:
        return asdict(self)



from dataclasses import dataclass
import math
import matplotlib.pyplot as plt
import pandas as pd
from .method import Method


@dataclass(frozen=True)
class Stochastics(Method):
    """
    買いのシグナルを計算で求める

    * %K・%D共に20％以下の時に、%Kが%Dを下から上抜いた時
    * %D・スロー%D共に20％以下の時に、%Dがスロー%Dを下から上抜いた時

    See Also:
        * https://www.moneypartners.co.jp/support/tech/sct.html

    """

    method_name: str = "stochastics"

    def _method(self, _df: pd.DataFrame) -> pd.DataFrame:

        _df["close"] = _df["close"].astype(float)
        _df["low"] = _df["low"].astype(float)
        _df["high"] = _df["high"].astype(float)
        _df["K"] = Stochastics._fast_stochastic_k(_df["close"], _df["low"], _df["high"], 9)
        _df["D"] = Stochastics._fast_stochastic_d(_df["K"])
        _df["SD"] = Stochastics._slow_stochastic_d(_df["D"])
        return _df

    def _signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        """
        買いと売りに関する指標を算出する
        """
        _df = _df.assign(
            shift_K=lambda x: x["K"].shift(1),
            shift_D=lambda x: x["D"].shift(1),
            shift_SD=lambda x: x["SD"].shift(1),
        )

        # 複数引数は関数を利用することで吸収
        _df["buy_signal"] = _df.apply(self._buy_signal_index_internal, axis=1)
        _df["sell_signal"] = _df.apply(self._sell_signal_index_internal, axis=1)
        return _df

    @staticmethod
    def _fast_stochastic_k(close, low, high, n):
        return (
            (close - low.rolling(window=n, center=False).min())
            / (high.rolling(window=n, center=False).max() - low.rolling(window=n, center=False).min())
        ) * 100

    @staticmethod
    def _fast_stochastic_d(stochastic_k):
        # ストキャスの%Dを計算（%Kの3日SMA）
        return stochastic_k.rolling(window=3, center=False).mean()

    @staticmethod
    def _slow_stochastic_d(stochastic_d):
        # ストキャスの%SDを計算（%Dの3日SMA）
        return stochastic_d.rolling(window=3, center=False).mean()

    @staticmethod
    def _buy_signal_index_internal(x: pd.Series) -> float:
        return Stochastics._buy_signal_index(x["K"], x["D"], x["SD"], x["shift_K"], x["shift_D"], x["shift_SD"])

    @staticmethod
    def _buy_signal_index(current_k, current_d, current_sd, prev_k, prev_d, prev_sd) -> float:
        if (current_k > 30) | (current_d > 30) | (current_sd > 30):
            return 0

        # %K・%D共に20％以下の時に、%Kが%Dを下から上抜いた時
        if current_k < 20 and current_d < 20:
            if (prev_d > prev_k) and (current_d < current_k):
                return current_k - current_d

        # %D・スロー%D共に20％以下の時に、%Dがスロー%Dを下から上抜いた時
        if current_d < 20 and current_sd < 20:
            if (prev_sd > prev_d) and (current_sd < current_d):
                return current_d - current_sd
        return 1 / math.exp(
            math.pow(current_k - 20, 2) / 100 + math.pow(current_d - 20, 2) / 100 + math.pow(current_sd - 20, 2) / 100
        )

    @staticmethod
    def _sell_signal_index_internal(x: pd.Series) -> float:
        return Stochastics._sell_signal_index(x["K"], x["D"], x["SD"], x["shift_K"], x["shift_D"], x["shift_SD"])

    @staticmethod
    def _sell_signal_index(current_k, current_d, current_sd, prev_k, prev_d, prev_sd) -> float:
        if (current_k < 70) | (current_d < 70) | (current_sd < 70):
            return 0

        # %K・%D共に80％以上の時に、%Kが%Dを上から下抜いた時
        if current_k > 80 and current_d > 80:
            if (prev_d < prev_k) and (current_d > current_k):
                return current_d - current_k

        # %D・スロー%D共に80％以上の時に、%Dがスロー%Dを上から下抜いた時
        # %D・スロー%D共に20％以下の時に、%Dがスロー%Dを下から上抜いた時
        if current_d > 80 and current_sd > 80:
            if (prev_sd < prev_d) and (current_sd > current_d):
                return current_d - current_sd
        return 1 / math.exp(
            math.pow(current_k - 20, 2) / 100 + math.pow(current_d - 20, 2) / 100 + math.pow(current_sd - 20, 2) / 100
        )

    def _visualize(self, _df: pd.DataFrame):
        fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, gridspec_kw={"height_ratios": [3, 1]}, figsize=(6, 5))
        # x軸のオートフォーマット
        fig.autofmt_xdate()

        # set candlestick
        self.add_ax_candlestick(ax1, _df)

        # plot macd
        ax2.plot(_df.index, _df["D"], label="%D")
        ax2.plot(_df.index, _df["SD"], label="%SD")
        ax2.legend(loc="center left")  # 各線のラベルを表示

        ax1.legend(loc="best")  # 各線のラベルを表示
        return fig



import pandas as pd
from kabutobashi.domain.method import Method
from typing import Union

# http://www.kabuciao.com/tech/deki/


def analysis_with(stock_df: pd.DataFrame, method: Method) -> pd.DataFrame:
    """

    Args:
        stock_df:
        method:

    Returns:
        pd.DataFrame
    """
    return stock_df.pipe(method)


def get_impact_with(stock_df: pd.DataFrame, method: Union[Method, list], **kwargs) -> dict:
    """
    入力のDataFrameに対して、分析の結果を返す関数

    Args:
        stock_df:
        method: 分析対象の手法
        **kwargs:

    Returns:
        Dict[str, float]

    Examples:
        >>> import kabutobashi as kb
        >>> get_impact_with(stock_df, [kb.SMA, kb.MACD])
        {"sma": 0.4, "macd": -0.04}
        >>> sma = kb.SMA(short_term=3, medium_term=15, long_term=50)
        >>> get_impact_with(stock_df, [sma, kb.MACD])
        {"sma": 0.2, "macd": -0.04}
    """
    # methodのpipeで渡す際の引数、impactにtrueを渡して直近の各手法の買い・売りの傾向を取得する
    kwargs.update({"impact": "true"})

    # 分析のリスト
    method_list = []
    if type(method) is list:
        method_list.extend(method)
    else:
        method_list.append(method)

    # 結果を格納するdict
    return {str(m): stock_df.pipe(m, **kwargs) for m in method_list}
